"""Defines the Docstring class, which interprets the AST."""

from collections import defaultdict
from typing import (
    Dict,
    List,
    Set,
    Union,
    Tuple,
    Optional,
)
from .node import (
    Node,
    NodeType,
)
from .parse import parse
from .lex import lex
from .peaker import Peaker


class Docstring(object):
    """The docstring class interprets the AST of a docstring."""

    def __init__(self, root):
        # type: (Union[Node, str]) -> None
        """Create a new docstring from the AST.

        Args:
            root: The root of the AST, or the docstring
                (as a string.)  If it is a string, the
                string will be parsed.

        """
        if isinstance(root, Node):
            self.root = root
        else:
            self.root = parse(Peaker(lex(root), lookahead=3))
        self._lookup = self._discover()

    def _discover(self):
        # type: () -> Dict[NodeType, List[Node]]
        """Walk the tree, finding all non-terminal nodes.

        Returns:
            A lookup table for compound Nodes by their NodeType.

        """
        lookup = defaultdict(list)  # type: Dict[NodeType, List[Node]]
        for node in self.root.breadth_first_walk(leaves=False):
            lookup[node.node_type].append(node)
        return lookup

    def has_short_description(self):
        # type: () -> bool
        """Tell if the docstring has a short description.

        Returns:
            True if the docstring has a short description.

        """
        return NodeType.SHORT_DESCRIPTION in self._lookup

    def has_long_description(self):
        # type: () -> bool
        """Tell if the docstring has a long description.

        Returns:
            True if the docstring has a long description.

        """
        return NodeType.LONG_DESCRIPTION in self._lookup

    def has_args_section(self):
        # type: () -> bool
        """Tell if the docstring has a args section.

        Returns:
            True if the docstring has a args section.

        """
        return NodeType.ARGS_SECTION in self._lookup

    def has_raises_section(self):
        # type: () -> bool
        """Tell if the docstring has a has raises section.

        Returns:
            True if the docstring has a has raises section.

        """
        return NodeType.RAISES_SECTION in self._lookup

    def has_yields_section(self):
        # type: () -> bool
        """Tell if the docstring has a has yields section.

        Returns:
            True if the docstring has a has yields section.

        """
        return NodeType.YIELDS_SECTION in self._lookup

    def has_returns_section(self):
        # type: () -> bool
        """Tell if the docstring has a has returns section.

        Returns:
            True if the docstring has a has returns section.

        """
        return NodeType.RETURNS_SECTION in self._lookup

    def get_return_type(self):
        # type: () -> Union[str, None]
        """Get the return type specified by the docstring, if any.

        Returns:
            The return type or None.

        """
        for return_node in self._lookup[NodeType.RETURNS_SECTION]:
            for node in return_node.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.TYPE:
                    type_repr = node.reconstruct_string()
                    if type_repr.startswith('(') and type_repr.endswith(')'):
                        type_repr = type_repr[2:-2]
                    return type_repr
        return None

    def get_exception_types(self):
        # type: () -> List[str]
        """Get the exception types described by the docstring.

        Returns:
            The types of exceptions described by the docstring.

        """
        ret = list()  # type: List[str]
        for raises_node in self._lookup[NodeType.RAISES_SECTION]:
            for node in raises_node.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.ITEM_NAME:
                    exception_name = node.children[0].value
                    ret.append(exception_name)
        return ret

    def get_yield_type(self):
        # type: () -> Union[str, None]
        """Get the yield type specified by the docstring, if any.

        Returns:
            The yield type or None

        """
        for yield_node in self._lookup[NodeType.YIELDS_SECTION]:
            for node in yield_node.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.TYPE:
                    type_repr = node.reconstruct_string()
                    if type_repr.startswith('(') and type_repr.endswith(')'):
                        type_repr = type_repr[2:-2]
                    return type_repr
        return None

    def get_argument_types(self):
        # type: () -> Dict[str, str]
        """Get a dictionary mapping arguments to types.

        Returns:
            A dictionary matching arguments to types.

        """
        argtypes = dict()  # type: Dict[str, str]
        item_names = list()  # type: List[Node]

        for arg_section in self._lookup[NodeType.ARGS_SECTION]:
            for node in arg_section.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.ITEM_NAME:
                    item_names.append(node)

        for item_name in item_names:
            argument = item_name.children[0]
            if len(item_name.children) > 1:
                type_node = item_name.children[1]
                type_repr = type_node.reconstruct_string()
                if type_repr.startswith('(') and type_repr.endswith(')'):
                    type_repr = type_repr[2:-2]
                argtypes[argument.value] = type_repr
            else:
                argtypes[argument.value] = None

        return argtypes

    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        """Get a map of the errors ignored to their targets.

        Returns:
            A dictionary containing the errors to ignore as keys and
            a list of which targets to apply these exceptions to as
            the values.  A blank list implies a global noqa.

        """
        encountered = set()  # type: Set[Node]
        global_noqas = set()  # type: Set[Node]
        noqas = defaultdict(list)  # type: Dict[str, List[str]]

        # Get exceptions with implied targets
        for item_node in self._lookup[NodeType.ITEM]:
            item = None
            for node in item_node.breadth_first_walk(leaves=False):
                # We will always encounter the item name first.
                if node.node_type == NodeType.ITEM_NAME:
                    item = node.children[0].value
                elif node.node_type == NodeType.NOQA_BODY:
                    exception = node.children[0]
                    encountered.add(exception)
                    noqas[exception.value].append(item)

        # Get all other exceptions
        for noqa_node in self._lookup[NodeType.NOQA_BODY]:
            exception = noqa_node.children[0]
            if exception in encountered:
                continue

            if len(noqa_node.children) == 1:
                global_noqas.add(exception)
                continue

            for word_node in noqa_node.children[1].children:
                word = word_node.value
                if word.endswith(','):
                    word = word[:-1]
                noqas[exception.value].append(word)

        # We overwrite any previous targets, because it was defined
        # as a global. (This could happen before a target is defined.)
        for global_noqa in global_noqas:
            noqas[global_noqa.value] = list()

        return dict(noqas)

    def _get_description(self, node_type):
        # type: (NodeType) -> str
        nodes = self._lookup[node_type]
        if not nodes:
            return None
        return ''.join([x.reconstruct_string() for x in nodes])

    def get_line_numbers(self, node_type):
        # type: (NodeType) -> Optional[Tuple[int, int]]
        """Get the line numbers for the first instance of the given section.

        Args:
            node_type: The NodeType which we want line numbers for.
                These should be unique instances. (I.e. they should be
                in the set of compound NodeTypes which only occur
                once in a docstring. For example, "Raises" and "Args".

        Returns:
            The line numbers for the first instance of the given node type.

        """
        nodes = self._lookup[node_type]
        if nodes:
            return nodes[0].line_numbers
        return None

    def get_line_numbers_for_value(self, node_type, value):
        # type: (NodeType, str) -> Optional[Tuple[int, int]]
        """Get the line number for a node with the given value.

        Args:
            node_type: The compound node which should contain the
                node we are searching for.
            value: The value of the node.

        Returns:
            A list of line numbers for nodes which match the
            parameters.

        """
        line_numbers = list()  # type: List[Tuple[int, int]]
        nodes = self._lookup[node_type]
        for node in nodes:
            for child in node.walk():
                if child.value == value and child.line_numbers:
                    return child.line_numbers
        return None

    @property
    def raises_description(self):
        # type: () -> str
        """Get the raises section of the docstring.

        Returns:
            The raises section of the docstring or None.

        """
        return self._get_description(NodeType.RAISES_SECTION)

    @property
    def returns_description(self):
        # type: () -> str
        """Get the returns section of the docstring.

        Returns:
            The returns section of the docstring, as a String,
            or None.

        """
        return self._get_description(NodeType.RETURNS_SECTION)

    @property
    def yields_description(self):
        # type: () -> str
        """Get the yield ssection of the docstring.

        Returns:
            The yields section, if it exists.

        """
        return self._get_description(NodeType.YIELDS_SECTION)

    @property
    def arguments_description(self):
        # type: () -> str
        """Get the arguments section of the docstring.

        Returns:
            The arguments section of the docstring, as a string,
            or None.

        """
        return self._get_description(NodeType.ARGS_SECTION)

    @property
    def short_description(self):
        # type: () -> str
        """Get the short description of the docstring.

        Returns:
            The short description in the docstring, or None.

        """
        return self._get_description(NodeType.SHORT_DESCRIPTION)

    @property
    def long_description(self):
        # type: () -> str
        """Get the long description of the docstring.

        Returns:
            The long description in the docstring, or None.

        """
        return self._get_description(NodeType.LONG_DESCRIPTION)

    @property
    def ignore_all(self):
        # type: () -> bool
        """Return whether we should ignore everything in the docstring.

        This happens when there is a bare noqa in the docstring, or
        there is "noqa: *" in the docstring.

        Returns:
            True if we should ignore everything, otherwise false.

        """
        for node in self._lookup[NodeType.NOQA]:
            body = None
            for child in node.walk():
                if child.node_type == NodeType.NOQA_BODY:
                    body = child
                    break

            if body is None or any([x.value == '*' for x in body.children]):
                return True

        return False



from collections import defaultdict
from typing import (  # noqa: F401
    Any,
    Dict,
    List,
    Set,
    Union,
    Tuple,
    Optional,
)

from .base import (  # noqa: F401
    BaseDocstring,
    DocstringStyle,
    Sections,
)
from ..node import (
    Node,
    NodeType,
)
from ..config import (
    Strictness,
)
from ..parse import (
    google,
)
from ..lex import lex
from ..peaker import Peaker


class Docstring(BaseDocstring):
    """The docstring class interprets the AST of a docstring."""

    def __init__(self, root, style=DocstringStyle.GOOGLE):
        # type: (Union[Node, str], DocstringStyle) -> None
        """Create a new docstring from the AST.

        Args:
            root: The root of the AST, or the docstring
                (as a string.)  If it is a string, the
                string will be parsed.
            style: The style of the docstring.  Discarded,
                since this Docstring is always the Google style.

        """
        if isinstance(root, Node):
            self.root = root
        else:
            self.root = google.parse(Peaker(lex(root), lookahead=3))
        self._lookup = self._discover()

    def _discover(self):
        # type: () -> Dict[NodeType, List[Node]]
        """Walk the tree, finding all non-terminal nodes.

        Returns:
            A lookup table for compound Nodes by their NodeType.

        """
        lookup = defaultdict(
            lambda: list()
        )  # type: Dict[NodeType, List[Node]]
        for node in self.root.breadth_first_walk(leaves=False):
            lookup[node.node_type].append(node)
        return lookup

    def get_section(self, section):
        # type: (Sections) -> Optional[str]
        nodes = []  # type: Optional[List[Node]]

        if section == Sections.SHORT_DESCRIPTION:
            nodes = self._lookup.get(NodeType.SHORT_DESCRIPTION, None)
        elif section == Sections.LONG_DESCRIPTION:
            nodes = self._lookup.get(NodeType.LONG_DESCRIPTION, None)
        elif section == Sections.ARGUMENTS_SECTION:
            nodes = self._lookup.get(NodeType.ARGS_SECTION, None)
        elif section == Sections.RAISES_SECTION:
            nodes = self._lookup.get(NodeType.RAISES_SECTION, None)
        elif section == Sections.YIELDS_SECTION:
            nodes = self._lookup.get(NodeType.YIELDS_SECTION, None)
        elif section == Sections.RETURNS_SECTION:
            nodes = self._lookup.get(NodeType.RETURNS_SECTION, None)
        elif section == Sections.NOQAS:
            nodes = self._lookup.get(NodeType.NOQA, None)
        else:
            raise Exception(
                'Unsupported section type {}'.format(section.name)
            )

        if not nodes:
            return None

        return_value = ''
        for node in nodes:
            return_value += '\n' + node.reconstruct_string()

        return return_value.strip() or None

    def _get_argument_types(self):
        # type: () ->  List[Optional[str]]
        """Get a list of types corresponding to arguments.

        Returns:
            A dictionary matching arguments to types.

        """
        lookup = self._get_compound_item_type_lookup(NodeType.ARGS_SECTION)
        if lookup is None:
            return []

        names_and_types = sorted(lookup.items())
        return [x[1] for x in names_and_types]

    def _get_return_type(self):
        # type: () -> Optional[str]
        """Get the return type specified by the docstring, if any.

        Returns:
            The return type or None.

        """
        for return_node in self._lookup[NodeType.RETURNS_SECTION]:
            for node in return_node.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.TYPE:
                    type_repr = node.reconstruct_string()
                    if type_repr.startswith('(') and type_repr.endswith(')'):
                        type_repr = type_repr[2:-2]
                    return type_repr
        return None

    def get_types(self, section):
        # type: (Sections) -> Union[None, str, List[Optional[str]]]
        if section == Sections.ARGUMENTS_SECTION:
            if NodeType.ARGS_SECTION not in self._lookup:
                return None
            return self._get_argument_types()
        elif section == Sections.RETURNS_SECTION:
            return self._get_return_type()
        elif section == Sections.YIELDS_SECTION:
            return self._get_yield_type()
        else:
            raise Exception(
                'Section type {} does not have types, '.format(
                    section.name
                ) + 'or is not yet supported'
            )
        return None

    def _get_compound_item_type_lookup(self, node_type):
        # type: (NodeType) -> Optional[Dict[str, Optional[str]]]
        """Get a map of names to types for the section.

        Args:
            node_type: The type of the section.

        Returns:
            A lookup of items to types.

        """
        if node_type not in self._lookup:
            return None

        item_types = dict()  # type: Dict[str, Optional[str]]
        for node in self._lookup[node_type]:
            for item_name in node.breadth_first_walk(leaves=False):
                if item_name.node_type != NodeType.ITEM_NAME:
                    continue
                name = item_name.first_instance(NodeType.WORD)
                if not name or not name.value:
                    continue
                _type = item_name.first_instance(NodeType.TYPE)
                if _type is None:
                    item_types[name.value] = None
                else:
                    _type_repr = _type.reconstruct_string()
                    _type_repr = _type_repr[1:-1].strip()
                    item_types[name.value] = _type_repr
        return item_types

    def _get_compound_items(self, node_type):
        # type: (NodeType) -> Optional[List[str]]
        lookup = self._get_compound_item_type_lookup(node_type)
        if lookup is None:
            return None
        return sorted(lookup.keys())

    def get_items(self, section):
        # type: (Sections) -> Optional[List[str]]
        if section == Sections.ARGUMENTS_SECTION:
            return self._get_compound_items(NodeType.ARGS_SECTION)
        elif section == Sections.RAISES_SECTION:
            return self._get_compound_items(NodeType.RAISES_SECTION)
        else:
            raise Exception(
                'Section type {} does not have items, '.format(
                    section.name
                ) + 'or is not yet supported.'
            )
        return None

    def _get_yield_type(self):
        # type: () -> Optional[str]
        """Get the yield type specified by the docstring, if any.

        Returns:
            The yield type or None

        """
        for yield_node in self._lookup[NodeType.YIELDS_SECTION]:
            for node in yield_node.breadth_first_walk(leaves=False):
                if node.node_type == NodeType.TYPE:
                    type_repr = node.reconstruct_string()
                    if type_repr.startswith('(') and type_repr.endswith(')'):
                        type_repr = type_repr[2:-2]
                    return type_repr
        return None

    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        """Get a map of the errors ignored to their targets.

        Returns:
            A dictionary containing the errors to ignore as keys and
            a list of which targets to apply these exceptions to as
            the values.  A blank list implies a global noqa.

        """
        encountered = set()  # type: Set[Node]
        global_noqas = set()  # type: Set[Node]
        noqas = defaultdict(list)  # type: Dict[str, List[str]]

        # Get exceptions with implied targets
        for item_node in self._lookup[NodeType.ITEM]:
            item = None  # type: Optional[str]
            for node in item_node.breadth_first_walk(leaves=False):
                # We will always encounter the item name first.
                if node.node_type == NodeType.ITEM_NAME:
                    assert node.children[0].value is not None
                    item = node.children[0].value
                elif node.node_type == NodeType.NOQA_BODY:
                    assert item is not None
                    exception = node.children[0]
                    assert exception.value is not None
                    encountered.add(exception)
                    noqas[exception.value].append(item)

        # Get all other exceptions
        for noqa_node in self._lookup[NodeType.NOQA_BODY]:
            exception = noqa_node.children[0]
            if exception in encountered:
                continue

            if len(noqa_node.children) == 1:
                global_noqas.add(exception)
                continue

            for word_node in noqa_node.children[1].children:
                word = word_node.value
                assert word is not None
                assert exception.value is not None
                if word.endswith(','):
                    word = word[:-1]
                noqas[exception.value].append(word)

        # We overwrite any previous targets, because it was defined
        # as a global. (This could happen before a target is defined.)
        for global_noqa in global_noqas:
            assert global_noqa.value is not None
            noqas[global_noqa.value] = list()

        return dict(noqas)

    def get_line_numbers(self, node_type):
        # type: (NodeType) -> Optional[Tuple[int, int]]
        """Get the line numbers for the first instance of the given section.

        Args:
            node_type: The NodeType which we want line numbers for.
                These should be unique instances. (I.e. they should be
                in the set of compound NodeTypes which only occur
                once in a docstring. For example, "Raises" and "Args".

        Returns:
            The line numbers for the first instance of the given node type.

        """
        nodes = self._lookup[node_type]
        if nodes:
            return nodes[0].line_numbers
        return None

    def get_line_numbers_for_value(self, node_type, value):
        # type: (NodeType, str) -> Optional[Tuple[int, int]]
        """Get the line number for a node with the given value.

        Args:
            node_type: The compound node which should contain the
                node we are searching for.
            value: The value of the node.

        Returns:
            A list of line numbers for nodes which match the
            parameters.

        """
        nodes = self._lookup[node_type]
        for node in nodes:
            for child in node.walk():
                if child.value == value and child.line_numbers:
                    return child.line_numbers
        return None

    @property
    def ignore_all(self):
        # type: () -> bool
        """Return whether we should ignore everything in the docstring.

        This happens when there is a bare noqa in the docstring, or
        there is "noqa: *" in the docstring.

        Returns:
            True if we should ignore everything, otherwise false.

        """
        for node in self._lookup[NodeType.NOQA]:
            body = None
            for child in node.walk():
                if child.node_type == NodeType.NOQA_BODY:
                    body = child
                    break

            if body is None or any([x.value == '*' for x in body.children]):
                return True

        return False

    def satisfies_strictness(self, strictness):
        # type: (Strictness) -> bool
        """Return true if the docstring has no more than the min strictness.

        Args:
            strictness: The minimum amount of strictness which should
                be present in the docstring.

        Returns:
            True if there is no more than the minimum amount of strictness.

        """
        if len(self.root.children) != 1:
            return False
        top_levels = {
            x.node_type for x in self.root.children[0].children
        }
        if strictness == Strictness.SHORT_DESCRIPTION:
            return top_levels == {NodeType.SHORT_DESCRIPTION}
        elif strictness == Strictness.LONG_DESCRIPTION:
            return top_levels in [
                {NodeType.SHORT_DESCRIPTION},
                {NodeType.SHORT_DESCRIPTION, NodeType.LONG_DESCRIPTION},
                # Shouldn't be possible, but if it is in the future, then
                # we should allow this.
                {NodeType.LONG_DESCRIPTION},
            ]
        else:
            # We don't need to check FULL_DESCRIPTION because it's the
            # same situation as when strictness is not satisfied.
            return False



"""The entry point for flake8."""

import ast  # noqa
from typing import (  # noqa
    Iterator,
    Tuple,
)

from .function_description import (
    get_function_descriptions,
)
from .integrity_checker import IntegrityChecker
from .config import (
    get_config,
)


__version__ = '0.4.1'


class DarglintChecker(object):

    name = 'flake8-darglint'
    version = __version__

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename
        self.verbosity = 2
        self.config = get_config()

    def run(self):
        # type: () -> Iterator[Tuple[int, int, str, type]]

        # Remember the last line number, so that if there is an
        # exception raised by Darglint, we can at least give a decent
        # idea of where it was raised.
        last_line = 1
        try:
            functions = get_function_descriptions(self.tree)
            checker = IntegrityChecker(
                self.config,
                raise_errors=False,
            )
            for function in functions:
                checker.run_checks(function)

            error_report = checker.get_error_report(
                self.verbosity,
                self.filename
            )
            for line, col, msg in error_report.flake8_report():
                last_line = line
                yield (line, col, msg, type(self))

        except Exception as ex:
            yield (last_line, 0, str(ex), type(self))


class StyleDarglintChecker(DarglintChecker):

    def run(self):
        for line, col, msg, selftype in super(StyleDarglintChecker, self).run():
            if msg.startswith('S'):
                yield line, col, msg, selftype


class InterfaceDarglintChecker(DarglintChecker):

    def run(self):
        for line, col, msg, selftype in super(InterfaceDarglintChecker, self).run():
            if msg.startswith('I'):
                yield line, col, msg, selftype



from typing import Union  # noqa

from .base import BaseDocstring  # noqa
from . import google, sphinx

from ..node import Node  # noqa


class Docstring(object):
    """A factory method for creating docstrings."""

    @staticmethod
    def from_google(root):
        # type: (Union[Node, str]) -> BaseDocstring
        return google.Docstring(root)

    @staticmethod
    def from_sphinx(root):
        # type: (Union[Node, str]) -> BaseDocstring
        return sphinx.Docstring(root)



"""Defines a function for lexing a comment, `lex`."""

from typing import (
    Iterator,
    List,
    Optional,
)
from .peaker import Peaker
from .token import Token, TokenType

# These convenience functions take an optional string
# because the peaker could return None when at the end
# of the stream.


def _is_space(char):
    # type: (Optional[str]) -> bool
    return char == ' '


def _is_newline(char):
    # type: (Optional[str]) -> bool
    return char == '\n'


def _is_colon(char):
    # type: (Optional[str]) -> bool
    return char == ':'


def _is_hash(char):
    # type: (Optional[str]) -> bool
    return char == '#'


def _is_separator(char):
    # type: (Optional[str]) -> bool
    """Check whether if `char` is a separator other than newline or space.

    Args:
        char: The character to check.

    Returns:
        true if `char` is a separator other than newline or space.

    """
    if char is None:
        return False
    return char.isspace() and not (_is_space(char) or _is_newline(char))


def _is_double_quotation(char):
    # type: (Optional[str]) -> bool
    return char == '"'


def _is_lparen(char):
    # type: (Optional[str]) -> bool
    return char == '('


def _is_rparen(char):
    # type: (Optional[str]) -> bool
    return char == ')'


def _is_word(char):
    # type: (str) -> bool
    return not any([
        _is_space(char),
        _is_newline(char),
        _is_colon(char),
        _is_separator(char),
        _is_double_quotation(char),
        _is_hash(char),
        _is_lparen(char),
        _is_rparen(char),
    ])


def lex(program):
    # type: (str) -> Iterator[Token]
    """Create a stream of tokens from the string.

    Args:
        program: The program to lex, as a string.

    Yields:
        Tokens lexed from the string.

    """
    extra = ''  # Extra characters which are pulled but unused from a check.
    peaker = Peaker((x for x in program or []))  # the stream
    line_number = 0
    while peaker.has_next():
        # Each of the following conditions must move the stream
        # forward and -- excepting separators -- yield a token.
        if _is_space(peaker.peak()):
            spaces = ''.join(peaker.take_while(_is_space))
            for _ in range(len(spaces) // 4):
                yield Token(' ' * 4, TokenType.INDENT, line_number)
        elif _is_newline(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.NEWLINE, line_number)
            line_number += 1
        elif _is_colon(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.COLON, line_number)
        elif _is_separator(peaker.peak()):
            peaker.take_while(_is_separator)
        elif _is_double_quotation(peaker.peak()):
            value = ''.join(peaker.take_while(_is_double_quotation))
            if len(value) >= 3:
                for _ in range(len(value) // 3):
                    yield Token('"""', TokenType.DOCTERM, line_number)
            else:
                extra = value
        elif _is_hash(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.HASH, line_number)
        elif _is_lparen(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.LPAREN, line_number)
        elif _is_rparen(peaker.peak()):
            value = peaker.next()
            yield Token(value, TokenType.RPAREN, line_number)
        else:
            value = ''.join(peaker.take_while(_is_word))
            if extra != '':
                value = extra + value
                extra = ''
            assert len(value) > 0, "There should be non-special characters."
            yield Token(value, TokenType.WORD, line_number)



"""A linter for docstrings following the google docstring format."""
import ast
from collections import deque
import sys
from typing import (
    Callable,
    Iterator,
    List,
    Set,
    Tuple,
    Optional,
    Union,
)

from .config import get_logger


logger = get_logger()


FunctionDef = ast.FunctionDef
if hasattr(ast, 'AsyncFunctionDef'):
    FunctionDef = (ast.FunctionDef, ast.AsyncFunctionDef)


def read_program(filename):  # type: (str) -> str
    """Read a program from a file.

    Args:
        filename: The name of the file to read. If set to '-', then we will
            read from stdin.

    Returns:
        The program as a single string.

    """
    program = None  # type: Optional[str]
    if filename == '-':
        program = sys.stdin.read()
    else:
        with open(filename, 'r') as fin:
            program = fin.read()
    return program


def _get_arguments(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[List[str], List[str]]
    arguments = list()  # type: List[str]
    types = list()  # type: List[str]

    def add_arg_by_name(name, arg):
        arguments.append(name)
        if arg.annotation is not None and hasattr(arg.annotation, 'id'):
            types.append(arg.annotation.id)
        else:
            types.append(None)

    for arg in fn.args.args:
        add_arg_by_name(arg.arg, arg)

    for arg in fn.args.kwonlyargs:
        add_arg_by_name(arg.arg, arg)

    # Handle single-star arguments.
    if fn.args.vararg is not None:
        name = '*' + fn.args.vararg.arg
        add_arg_by_name(name, fn.args.vararg)

    if fn.args.kwarg is not None:
        name = '**' + fn.args.kwarg.arg
        add_arg_by_name(name, fn.args.kwarg)

    return arguments, types


def _walk(fun, skip):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef], Callable) -> Iterator[ast.AST]
    """Walk through the nodes in this function, skipping as necessary.

    ast.walk goes through nodes in an arbitrary order, and doesn't
    allow you to skip an entire branch (as far as I know.)  This
    function will perform an in-order breadth-first traversal, and will
    skip unnecessary branches.

    Args:
        fun: The function to walk.
        skip: A function which returns True if we should skip the current
            node and all of its children.

    Yields:
        Children of the function and the function itself.

    """
    queue = deque()  # type: deque
    queue.appendleft(fun)
    while len(queue) > 0:
        curr = queue.pop()
        if skip(curr):
            continue
        if hasattr(curr, 'body'):
            queue.extendleft(curr.body)
        if hasattr(curr, 'handlers'):
            queue.extendleft(curr.handlers)
        yield curr


def _has_return(fun):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool
    """Return true if the function has a fruitful return.

    Args:
        fun: A function node to check.

    Returns:
        True if there is a fruitful return, otherwise False.

    """
    def skip(f):
        return f != fun and isinstance(f, FunctionDef)

    for node in _walk(fun, skip):
        if isinstance(node, ast.Return) and node.value is not None:
            return True
    return False


def _has_yield(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool
    for node in ast.walk(fun):
        if isinstance(node, ast.Yield) or isinstance(node, ast.YieldFrom):
            return True
    return False


def _get_docstring(fun):  # type: (ast.AST) -> str
    return ast.get_docstring(fun)


def _get_all_functions(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
    for node in ast.walk(tree):
        if isinstance(node, FunctionDef):
            yield node


def _get_all_classes(tree):  # type: (ast.AST) -> Iterator[ast.ClassDef]
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef):
            yield node


def _get_all_methods(tree):  # type: (ast.AST) -> Iterator[Union[ast.FunctionDef, ast.AsyncFunctionDef]]
    for klass in _get_all_classes(tree):
        for fun in _get_all_functions(klass):
            yield fun


def _get_decorator_names(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> List[str]
    """Get decorator names from the function.

    Args:
        fun: The function whose decorators we are getting.

    Returns:
        The names of the decorators. Does not include setters and
        getters.

    """
    ret = list()
    for decorator in fun.decorator_list:
        # Attributes (setters and getters) won't have an id.
        if hasattr(decorator, 'id'):
            ret.append(decorator.id)
    return ret


def _is_classmethod(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool
    return 'classmethod' in _get_decorator_names(fun)


def _is_staticmethod(fun):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> bool
    return 'staticmethod' in _get_decorator_names(fun)


def _get_stripped_method_args(method):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Tuple[List[str], List[str]]
    args, types = _get_arguments(method)
    if 'cls' in args and _is_classmethod(method):
        args.remove('cls')
        types.pop(0)
    elif 'self' in args and not _is_staticmethod(method):
        args.remove('self')
        types.pop(0)
    return args, types


def _get_all_raises(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Iterator[ast.Raise]
    for node in ast.walk(fn):
        if isinstance(node, ast.Raise):
            yield node


def _get_exception_name(raises):  # type: (ast.Raise) -> str
    if isinstance(raises.exc, ast.Name):
        return raises.exc.id
    elif isinstance(raises.exc, ast.Call):
        if hasattr(raises.exc.func, 'id'):
            return raises.exc.func.id
        elif hasattr(raises.exc.func, 'attr'):
            return raises.exc.func.attr
        else:
            logger.error(
                'Raises function call has neither id nor attr.'
                'has only: %s' % str(dir(raises.exc.func))
            )
    else:
        raise Exception('Unexpected type in raises expression: {}'.format(
            type(raises.exc)
        ))
    return ''


def _get_exceptions_raised(fn):  # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Set[str]
    ret = set()  # type: Set[str]
    for raises in _get_all_raises(fn):
        # TODO: Handle this?
        # There is a bare raise in the function, no type given.
        if raises.exc is None:
            continue
        ret.add(_get_exception_name(raises))
    return ret


def _get_return_type(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Optional[str]
    if fn.returns is not None and hasattr(fn.returns, 'id'):
        return fn.returns.id
    return None


def _get_all_variable_names(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> Iterator[str]
    """Get all variable names used in the function.

    Args:
        fn: The function.

    Yields:
        The nodes which represent names.

    """
    for node in ast.walk(fn):
        if isinstance(node, ast.Name):
            yield node.id


def get_line_number_from_function(fn):
    # type: (Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> int
    """Get the line number for the end of the function signature.

    The function signature can be farther down when the parameter
    list is split across multiple lines.

    Args:
        fn: The function from which we are getting the line number.

    Returns:
        The line number for the start of the docstring for this
        function.

    """
    line_number = fn.lineno
    if hasattr(fn, 'args') and fn.args.args:
        last_arg = fn.args.args[-1]
        line_number = last_arg.lineno
    return line_number


class FunctionDescription(object):
    """Describes a function or method.

    Whereas a `Docstring` object describes a function's docstring,
    a `FunctionDescription` describes the function itself.  (What,
    ideally, the docstring should describe.)

    """

    def __init__(self, is_method, function):
        # type: (bool, Union[ast.FunctionDef, ast.AsyncFunctionDef]) -> None
        """Create a new FunctionDescription.

        Args:
            is_method: True if this is a method. Will attempt to remove
                self or cls if appropriate.
            function: The base node of the function.

        """
        self.is_method = is_method
        self.function = function
        self.line_number = get_line_number_from_function(function)
        self.name = function.name
        if is_method:
            self.argument_names, self.argument_types = (
                _get_stripped_method_args(function)
            )
        else:
            self.argument_names, self.argument_types = _get_arguments(function)
        self.has_return = _has_return(function)
        self.return_type = _get_return_type(function)
        self.has_yield = _has_yield(function)
        self.docstring = _get_docstring(function)
        try:
            self.raises = _get_exceptions_raised(function)
        except Exception as ex:
            msg = '{}: {}'.format(self.name, ex)
            logger.error(msg)
            raise
        self.variables = _get_all_variable_names(function)


def get_function_descriptions(program):
    # type: (ast.AST) -> List[FunctionDescription]
    """Get function name, args, return presence and docstrings.

    This function should be called on the top level of the
    document (for functions), and on classes (for methods.)

    Args:
        program: The tree representing the entire program.

    Returns:
        A list of function descriptions pulled from the ast.

    """
    ret = list()  # type: List[FunctionDescription]

    methods = set(_get_all_methods(program))
    for method in methods:
        ret.append(FunctionDescription(is_method=True, function=method))

    functions = set(_get_all_functions(program)) - methods
    for function in functions:
        ret.append(FunctionDescription(is_method=False, function=function))

    return ret



r"""A parser for Sphinx-style docstrings.

The EBNF for the parser is as follows:

  docstring = short-description, [long-description];
            | short-description
                , [long-description]
                , item
                , {[newline], item}
                , newline;

  short-description = unline
                    | line, newline;
  long-description = line, {line}, newline;

  item = indent, item-name, item-definition;
  section-name = colon, keyword, [word], colon;
  item-definition = line, {line};
  line = unline, newline
  unline = { word
           , hash
           , colon
           , indent
           , keyword
           , lparen
           , rparen
           }, [noqa];

  noqa = noqa-head, [colon, list]
  noqa-head = hash, noqa-keyword
  list = word {"," word}

  keyword = "arg"
          | "argument"
          | "param"
          | "parameter"
          | "key"
          | "keyword"
          | "raises"
          | "raise"
          | "except"
          | "exception"
          | "var"
          | "ivar"
          | "cvar"
          | "returns"
          | "return"
          | "yield"
          | "yields"
          | "type"
          | "vartype"
          | "rtype"
          | "ytype";

  indent  = " " * 4;
  word    = ? r/[^\ \n\:\"\t]+/ ?;
  noqa-keyword = "noqa"
  hash = "#"
  lparen = "("
  rparen = ")"
  colon   = ":";
  newline = "\n";

"""

from itertools import chain
from typing import (  # noqa
    Dict,
    List,
)

from ..node import (
    Node,
    NodeType,
)
from ..peaker import Peaker  # noqa
from ..token import (  # noqa
    Token,
    TokenType,
)
from .common import (
    Assert,
    AssertNotEmpty,
    ParserException,
    parse_colon,
    parse_hash,
    parse_indent,
    parse_keyword,
    parse_lparen,
    parse_noqa,
    parse_rparen,
    parse_word,
)

KEYWORDS = {
    'arg': NodeType.ARGUMENTS,
    'argument': NodeType.ARGUMENTS,
    'param': NodeType.ARGUMENTS,
    'parameter': NodeType.ARGUMENTS,
    'key': NodeType.ARGUMENTS,
    'keyword': NodeType.ARGUMENTS,
    'type': NodeType.TYPE,
    'raises': NodeType.RAISES,
    'raise': NodeType.RAISES,
    'except': NodeType.RAISES,
    'exception': NodeType.RAISES,
    'var': NodeType.VARIABLES,
    'ivar': NodeType.VARIABLES,
    'cvar': NodeType.VARIABLES,
    'vartype': NodeType.TYPE,
    'returns': NodeType.RETURNS,
    'return': NodeType.RETURNS,
    'rtype': NodeType.TYPE,
    'yield': NodeType.YIELDS,
    'yields': NodeType.YIELDS,
    'ytype': NodeType.TYPE,
}

_KEYWORD_TO_SECTION = {
    NodeType.ARGUMENTS: NodeType.ARGS_SECTION,
    NodeType.RAISES: NodeType.RAISES_SECTION,
    NodeType.VARIABLES: NodeType.VARIABLES_SECTION,
    NodeType.RETURNS: NodeType.RETURNS_SECTION,
    NodeType.YIELDS: NodeType.YIELDS_SECTION,
}


def _in_keywords(peaker, offset=1):
    # type: (Peaker[Token], int) -> bool
    token = peaker.peak(offset)
    if token is None:
        return False
    return token.value in KEYWORDS


def _is(expected_type, peaker, offset=1):
    # type: (TokenType, Peaker[Token], int) -> bool
    """Check if the peaker's next value is the given type.

    Args:
        expected_type: The type we're checking.
        peaker: The peaker.
        offset: The lookahead to use.  (Most of the time, this
            will be 1 -- the current token.)

    Returns:
        True if the next token is the given type, false
        otherwise. (Including if there are no more tokens.)

    """
    token = peaker.peak(offset)
    if token is not None:
        return token.token_type == expected_type
    return False


def parse_line(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'Parsing line.')
    children = []  # type: List[Node]

    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker):
        if _is(TokenType.WORD, peaker) and _in_keywords(peaker):
            children.append(parse_keyword(peaker, KEYWORDS))
        elif _is(TokenType.WORD, peaker):
            children.append(parse_word(peaker))
        elif _is(TokenType.COLON, peaker):
            children.append(parse_colon(peaker))
        elif _is(TokenType.INDENT, peaker):
            children.append(parse_indent(peaker))
        elif _is(TokenType.LPAREN, peaker):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, peaker):
            children.append(parse_rparen(peaker))
        elif _is(TokenType.HASH, peaker):
            token = peaker.peak(lookahead=2)
            if token is not None and token.value == 'noqa':
                children.append(parse_noqa(peaker))
            else:
                children.append(parse_hash(peaker))
        else:
            token = peaker.peak()
            assert token is not None
            raise ParserException('Unexpected type {} in line.'.format(
                token.token_type
            ))

    # It is possible that there are no children at this point, in
    # which case there is likely just the newline.  In this case,
    # we try to set the token so that the line can have a line
    # number.
    token = None
    if peaker.has_next():
        if not children:
            token = peaker.next()
        else:
            peaker.next()  # Throw away newline.
    return Node(
        NodeType.LINE,
        children=children,
        token=token,
    )


def parse_short_description(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse short description')
    Assert(
        not _is(TokenType.NEWLINE, peaker),
        'Must have short description in docstring.'
    )
    return Node(
        NodeType.SHORT_DESCRIPTION,
        children=[parse_line(peaker)],
    )


def _at_item(peaker):
    # type: (Peaker[Token]) -> bool
    token = peaker.peak(lookahead=2)
    if token is None:
        return False
    return (_is(TokenType.COLON, peaker)
            and token.value in KEYWORDS)


def parse_long_description(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse long description')
    children = list()  # type: List[Node]
    while peaker.has_next() and not _at_item(peaker):
        children.append(parse_line(peaker))

    return Node(
        node_type=NodeType.LONG_DESCRIPTION,
        children=children,
    )


def parse_item_definition(peaker):
    # type: (Peaker[Token]) -> Node
    children = [parse_line(peaker)]
    while _is(TokenType.INDENT, peaker):
        children.append(parse_line(peaker))
    return Node(
        node_type=NodeType.ITEM_DEFINITION,
        children=children,
    )


def parse_item_head_with_argument(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse item')
    children = list()  # type: List[Node]

    token = peaker.peak()
    assert token is not None
    Assert(
        _is(TokenType.COLON, peaker),
        'Expected item to start with {} but was {}: {}'.format(
            TokenType.COLON, token.token_type, token.value
        ),
    )
    children.append(parse_colon(peaker))

    AssertNotEmpty(peaker, 'parse item')
    token = peaker.peak()
    assert token is not None
    Assert(
        _in_keywords(peaker),
        'Expected a keyword (e.g. "arg", "returns", etc.) but was {}'.format(
            token.value
        )
    )
    keyword = parse_keyword(peaker, KEYWORDS)
    children.append(keyword)

    Assert(
        not _is(TokenType.COLON, peaker),
        'Expected argument in {} section.'.format(
            children[-1].value
        ),
    )

    #           1     2  3
    # :param <type> <arg>:
    if not _is(TokenType.COLON, peaker, offset=2):
        _type = parse_word(peaker)
        children.append(Node(
            node_type=NodeType.TYPE,
            children=[_type],
        ))
    children.append(parse_word(peaker))

    AssertNotEmpty(peaker, 'parse item head end')
    token = peaker.peak()
    assert token is not None
    Assert(
        _is(TokenType.COLON, peaker),
        'Expected item head to end with {} but was {} {}'.format(
            TokenType.COLON,
            token.token_type,
            repr(token.value),
        ),
    )
    children.append(parse_colon(peaker))
    return Node(
        node_type=NodeType.ITEM_NAME,
        children=children,
    )


def parse_item_head_without_argument(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse item')
    children = list()  # type: List[Node]

    token = peaker.peak()
    assert token is not None
    Assert(
        _is(TokenType.COLON, peaker),
        'Expected item to start with {} but was {}: {}'.format(
            TokenType.COLON, token.token_type, token.value,
        ),
    )
    children.append(parse_colon(peaker))

    AssertNotEmpty(peaker, 'parse item')
    token = peaker.peak()
    assert token is not None
    Assert(
        token.value in {
            'returns', 'return', 'yields', 'yield', 'rtype', 'ytype'
        },
        'Expected section which doesn\'t take an argument '
        '({}) but was {}'.format(
            'returns, return, yields or yield', token.value
        )
    )
    keyword = parse_keyword(peaker, KEYWORDS)
    children.append(keyword)

    if not _is(TokenType.COLON, peaker):
        _type = parse_word(peaker)
        children.append(Node(
            node_type=NodeType.TYPE,
            children=[_type],
        ))

    AssertNotEmpty(peaker, 'parse item head end')
    token = peaker.peak()
    assert token is not None
    Assert(
        _is(TokenType.COLON, peaker),
        'Expected item head to end with {} but was {} {}'.format(
            TokenType.COLON,
            token.token_type,
            repr(token.value),
        ),
    )
    children.append(parse_colon(peaker))
    return Node(
        node_type=NodeType.ITEM_NAME,
        children=children,
    )


def parse_item_head(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse item')

    colon_token = peaker.peak()
    assert colon_token is not None
    Assert(
        _is(TokenType.COLON, peaker),
        'Expected item to start with {} but was {}: "{}". '
        'Are your indents four spaces?'.format(
            TokenType.COLON, colon_token.token_type, colon_token.value,
        ),
    )

    section_token = peaker.peak(lookahead=2)
    Assert(
        section_token is not None,
        'Stream unexpectedly empty while parsing item head.'
    )
    assert section_token is not None
    Assert(
        section_token.value in KEYWORDS,
        'Expected a keyword (e.g. "arg", "returns", etc.) but was {}'.format(
            section_token.value
        )
    )

    if section_token.value in {
            'returns', 'return', 'yield', 'yields', 'rtype', 'ytype'
    }:
        return parse_item_head_without_argument(peaker)
    else:
        return parse_item_head_with_argument(peaker)


def parse_item(peaker):
    # type: (Peaker[Token]) -> Node
    head = parse_item_head(peaker)

    keyword = head.children[1]

    if keyword.node_type == NodeType.TYPE:
        allowable_types = ['type', 'rtype', 'vartype', 'ytype']
        Assert(
            keyword.value in allowable_types,
            'Unable to determine section type from keyword {}: '
            'expected one of {}'.format(
                keyword.value,
                str(allowable_types),
            )
        )
        if keyword.value == 'rtype':
            section_type = NodeType.RETURNS_SECTION
        elif keyword.value == 'vartype':
            section_type = NodeType.VARIABLES_SECTION
        elif keyword.value == 'type':
            section_type = NodeType.ARGS_SECTION
        elif keyword.value == 'ytype':
            section_type = NodeType.YIELDS_SECTION
    else:
        section_type = _KEYWORD_TO_SECTION[keyword.node_type]

    children = [
        head,
        parse_item_definition(peaker),
    ]

    return Node(
        node_type=section_type,
        children=[Node(
            node_type=NodeType.ITEM,
            children=children,
        )]
    )


def parse_description(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse description')
    children = [
        parse_short_description(peaker),
    ]
    long_descriptions = list()  # type: List[Node]
    while peaker.has_next() and not _at_item(peaker):
        long_descriptions.append(
            parse_long_description(peaker)
        )

    # Consolidate all long descriptions into one.
    if long_descriptions:
        desc = [x.children for x in long_descriptions]
        children.append(
            Node(
                node_type=NodeType.LONG_DESCRIPTION,
                children=list(chain(*desc))
            )
        )

    return Node(
        node_type=NodeType.DESCRIPTION,
        children=children
    )


def parse(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse docstring')
    children = [
        parse_description(peaker)
    ]

    while peaker.has_next():
        children.append(parse_item(peaker))

        # Consume extra newlines.
        while _is(TokenType.NEWLINE, peaker):
            peaker.next()

    return Node(
        node_type=NodeType.DOCSTRING,
        children=children,
    )

    # TODO: In the parse function, parse everything,
    # then run over it and conglomerate the sections.
    # that will allow us to treat the tree the same as we
    # treat the Google tree.



"""Defines the command line interface for darglint."""
import argparse
import ast
import pathlib
import sys

from .function_description import (
    read_program,
    get_function_descriptions,
)
from .integrity_checker import IntegrityChecker
from .config import (
    Configuration,
    get_config,
    Strictness,
)
from .docstring.base import DocstringStyle


# ---------------------- ARGUMENT PARSER -----------------------------

parser = argparse.ArgumentParser(description='Check docstring validity.')
parser.add_argument(
    '--message-template',
    '-m',
    type=str,
    help=(
        'Specify a message template.  This is a Python format string '
        'describing errors, which can access the following attributes:\n'
        '    line: The line number,\n'
        '    msg: The error message,\n'
        '    msg_id: The error code,\n'
        '    obj: The function/method name,\n'
        '    path: The relative file path.\n'
    ),
)
parser.add_argument(
    '--raise-syntax',
    action='store_true',
    help=(
        'When a docstring is incorrectly formatted, raise an exception '
        'rather than storing the error.  Useful for debugging darglint.'
    ),
)
parser.add_argument(
    '--verbosity',
    '-v',
    default=1,
    type=int,
    choices=[1, 2],
    help='The level of verbosity.',
)
parser.add_argument(
    '--version',
    action='store_true',
    help=(
        'Return the current version number of darglint.'
    ),
)
parser.add_argument(
    'files',
    nargs='*',
    help=(
        'The python source files to check. If "-" is given, then stdin will '
        'be read.'
    ),
)
parser.add_argument(
    '--no-exit-code',
    '-x',
    action='store_true',
    help=(
        'Exit with status 0, even on errors.  By default, darglint '
        'exits with status 1 when errors are encountered.  Giving '
        'this flag prevents that.  Useful when invocating with xargs '
        'and you want to see all errors.  '
        'Ex: `find . -name "*.py" | xargs darglint -x`'
    ),
)
parser.add_argument(
    '--list-errors',
    action='store_true',
    help=(
        'Print a list of error codes and what they represent.'
    )
)
parser.add_argument(
    '-s',
    '--docstring-style',
    default=None,
    choices=['google', 'sphinx'],
    help=(
        'The docstring style used in the given project. Currently, '
        'only google or sphinx styles are supported.'
    )
)
parser.add_argument(
    '-z',
    '--strictness',
    default=None,
    choices=[
        'short',
        'long',
        'full',
    ],
    help=(
        'The minimum strictness when checking docstrings. '
        '`short`, for example, will result in one-line '
        'docstrings always being accepted.  Anything more than one line '
        'would go through the full check.'
    ),
)

# ---------------------- MAIN SCRIPT ---------------------------------


def get_error_report(filename,
                     verbosity,
                     config,
                     raise_errors_for_syntax,
                     message_template=None,
                     ):
    # type: (str, int, Configuration, bool, str) -> str
    """Get the error report for the given file.

    Args:
        filename: The name of the module to check.
        verbosity: The level of verbosity, in the range [1, 3].
        config: The configuration for the error report.
        raise_errors_for_syntax: True if we want parser errors
            to propagate up (crashing darglint.)  This is useful
            if we are developing on darglint -- we can get the stack
            trace and know exactly where darglint failed.
        message_template: A python format string for specifying
            how the message should appear to the user.

    Returns:
        An error report for the file.

    """
    program = read_program(filename)
    tree = ast.parse(program)
    functions = get_function_descriptions(tree)
    checker = IntegrityChecker(
        config,
        raise_errors=raise_errors_for_syntax,
    )
    for function in functions:
        checker.run_checks(function)
    return checker.get_error_report_string(
        verbosity,
        filename,
        message_template=message_template,
    )


def print_error_list():
    print('\n'.join([
        'I101: The docstring is missing a parameter in the definition.',
        'I102: The docstring contains a parameter not in function.',
        'I103: The docstring parameter type doesn\'t match function.',
        'I201: The docstring is missing a return from definition.',
        'I202: The docstring has a return not in definition.',
        'I203: The docstring parameter type doesn\'t match function.',
        'I301: The docstring is missing a yield present in definition.',
        'I302: The docstring has a yield not in definition.',
        'I401: The docstring is missing an exception raised.',
        'I402: The docstring describes an exception not explicitly raised.',
        'S001: Describes that something went wrong in parsing the docstring.',
        'S002: An argument/exception lacks a description.',
    ]))


def print_version():
    print('0.6.0')


def main():
    # type: () -> None
    """Run darglint.

    Called as a script when setup.py is installed.

    """
    args = parser.parse_args()
    exit_code = not args.no_exit_code
    encountered_errors = False

    if args.list_errors:
        print_error_list()
        sys.exit(0)

    if args.version:
        print_version()

    # Expand directories.
    files = []
    for f in args.files:
        p = pathlib.Path(f)
        if not p.is_dir():
            files.append(f)
        # Convert back to strings to not require modifications of any
        # subsequent code.
        files.extend(str(i) for i in p.glob('**/*.py'))

    try:
        config = get_config()

        if args.docstring_style == 'sphinx':
            config.style = DocstringStyle.SPHINX
        elif args.docstring_style == 'google':
            config.style = DocstringStyle.GOOGLE

        if args.strictness == 'short':
            config.strictness = Strictness.SHORT_DESCRIPTION
        elif args.strictness == 'long':
            config.strictness = Strictness.LONG_DESCRIPTION
        elif args.strictness == 'full':
            config.strictness = Strictness.FULL_DESCRIPTION

        raise_errors_for_syntax = args.raise_syntax or False
        for filename in files:
            error_report = get_error_report(
                filename,
                args.verbosity,
                config,
                raise_errors_for_syntax,
                message_template=args.message_template,
            )
            if error_report:
                print(error_report + '\n')
                encountered_errors = True
    except Exception as exc:
        # Exit with status 129 regardless of whether user wants a
        # exit code or not -- darglint failed, and it should
        # look like it failed.
        print(exc, file=sys.stderr)
        sys.exit(129)
    if encountered_errors and exit_code:
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == '__main__':
    main()



"""Defines the tokens that result from lexing, as well as their types."""

from enum import Enum


class TokenType(Enum):
    """The possible types of tokens."""

    WORD = 1
    COLON = 2
    DOCTERM = 3  # """
    NEWLINE = 4
    INDENT = 5  # Assumed to always be 4 spaces
    HASH = 6  # A hash mark (for comments/noqa).
    LPAREN = 7  # (
    RPAREN = 8  # )


class Token(object):
    """A token representing anything which can appear in a docstring."""

    def __init__(self,
                 value: str,
                 token_type: TokenType,
                 line_number: int
                 ) -> None:
        """Create a new Token.

        Args:
            value: The value of the token. (The actual string.)
            token_type: The type of token this represents.
            line_number: The line number where this token resides.
                Used when reporting errors.

        """
        self.value = value
        self.token_type = token_type
        self.line_number = line_number

    def __str__(self):
        """Return readable representation for debugging.

        Returns:
            A readable representation for debugging.

        """
        return '<Token {} {}>'.format(self.value, self.token_type)

    def __repr__(self):
        """Return readable representation for debugging.

        Returns:
            A readable representation for debugging.

        """
        return str(self)



"""Defines a node for the docstring AST."""

from collections import deque

from enum import Enum
from typing import (  # noqa: F401
    Any,
    List,
    Iterator,
    Tuple,
    Optional,
)
from .token import (  # noqa: F401
    Token,
)


class NodeType(Enum):
    """The type of node in a docstring."""
    # Non-Terminals
    DOCSTRING = 0
    SUMMARY = 1
    DESCRIPTION = 2
    SHORT_DESCRIPTION = 26
    TYPE = 10
    LINE = 11
    SECTION_HEAD = 13
    SECTION_SIMPLE_BODY = 14
    SECTION_COMPOUND_BODY = 15
    SECTION = 16
    ITEM_NAME = 17
    ITEM_DEFINITION = 18
    ITEM = 19
    ARGS_SECTION = 22
    RAISES_SECTION = 23
    RETURNS_SECTION = 24
    YIELDS_SECTION = 25
    LONG_DESCRIPTION = 27
    NOQA = 28
    LIST = 29
    NOQA_HEAD = 30
    NOQA_BODY = 31
    VARIABLES_SECTION = 34

    # Terminals
    KEYWORD = 3
    WORD = 4
    COLON = 5
    RETURNS = 6
    ARGUMENTS = 7
    YIELDS = 8
    RAISES = 9
    INDENT = 12
    LPAREN = 20
    RPAREN = 21
    HASH = 32
    VARIABLES = 33


class Node(object):
    """A node in a docstring AST."""

    def __init__(self, node_type, value=None, children=None, token=None):
        # type: (NodeType, str, List[Node], Token) -> None
        """Instantiate the new node.

        If the node is terminal, it will get the line number from the
        token.  If it is non-terminal, it will derive the line number(s)
        from the children.

        Args:
            node_type: The type of node.
            value: The value of the node.  Should only be specified for
                terminal nodes.
            children: The children of this node. Should only be specified
                for non-terminal nodes.
            token: The token this node was made from.  Should only be
                specified for terminal nodes.

        """
        self.node_type = node_type
        self.value = value
        self.children = children or list()
        self.line_numbers = None  # type: Optional[Tuple[int, int]]
        if token:
            self.line_numbers = (
                token.line_number,
                token.line_number
            )
        elif children:
            child_min_line_numbers = [
                x.line_numbers[0] for x in children
                if x.line_numbers
            ]
            child_max_line_numbers = [
                x.line_numbers[1] for x in children
                if x.line_numbers
            ]
            if child_min_line_numbers and child_max_line_numbers:
                self.line_numbers = (
                    min(child_min_line_numbers),
                    max(child_max_line_numbers),
                )

    @property
    def is_keyword(self):
        return self.node_type in {
            NodeType.KEYWORD,
            NodeType.RETURNS,
            NodeType.ARGUMENTS,
            NodeType.YIELDS,
            NodeType.RAISES,
        }

    @property
    def is_leaf(self):
        """Tell whether this node is a leaf.

        Returns:
            True if this node is a leaf.

        """
        return self.node_type in {
            NodeType.KEYWORD,
            NodeType.WORD,
            NodeType.COLON,
            NodeType.RETURNS,
            NodeType.ARGUMENTS,
            NodeType.YIELDS,
            NodeType.RAISES,
            NodeType.INDENT,
            NodeType.LPAREN,
            NodeType.RPAREN,
            NodeType.HASH,
        }

    def first_instance(self, node_type):
        # type: (NodeType) -> Optional[Node]
        for child in self.breadth_first_walk():
            if child.node_type == node_type:
                return child
        return None

    def walk(self):
        # type: () -> Iterator[Node]
        """Iterate over nodes in the tree rooted at this node.

        Recursion isn't too horrible here, as the syntax tree
        in this case never gets very deep.  Does a post-order
        traversal of the tree.

        Yields:
            Nodes in the tree.

        """
        for child in self.children:
            yield from child.walk()
        yield self

    def breadth_first_walk(self, leaves=True):
        # type: (bool) -> Iterator[Node]
        """Iterate over the nodes in the tree using Breadth First Traversal.

        A breadth-first traversal will be much faster when identifying
        sections.

        Args:
            leaves: True if leaves of the tree should also be yielded,
                otherwise only parent nodes will be yielded.

        Yields:
            Nodes in the tree.

        """
        queue = deque()  # type: deque
        queue.appendleft(self)
        while queue:
            curr = queue.pop()
            yield curr
            queue.extendleft([
                child for child in curr.children
                if leaves or not child.is_leaf
            ])

    def reconstruct_string(self):
        # type: () -> str
        """Attempt to reconstruct how the node looked before.

        Unfortunately, with how the parser works, we can't do this
        perfectly.  Nor would it likely be worthwhile. For example, if
        there are extra spaces (not in multiples of 4), then they will
        not be represented.

        If the parser ever changes, this will likely have to change.
        Which is fine, because it's gross anyway.

        Returns:
            Something close to the original string.

        """
        lines = [[]]  # type: List[List[str]]
        keyword = None
        for child in self.walk():
            if child.node_type == NodeType.INDENT:
                # When joined, it will have 4 spaces.
                lines[-1].append(' '*3)
            elif child.node_type == NodeType.COLON:
                # Remove the space before the colon. (This is more common.)
                if keyword:
                    lines[-1].append(keyword + ':')
                    keyword = None
                    lines.append(list())
                elif lines[-1]:
                    lines[-1][-1] += ':'
                elif child.value:
                    lines[-1].append(child.value)
            elif child.is_keyword:
                # Keywords always have colons after them, so wait for
                # the colon.
                keyword = child.value
            elif child.is_leaf and child.value:
                lines[-1].append(child.value)
            elif child.node_type == NodeType.LINE:
                lines.append(list())
        return '\n'.join([' '.join(line) for line in lines])



"""Common parsing utilities."""

from typing import (  # noqa: F401
    Any,
    Dict,
    Optional,
    Tuple,
)

from ..errors import (
    GenericSyntaxError,
)
from ..node import (
    Node,
    NodeType,
)
from ..peaker import (  # noqa: F401
    Peaker,
)
from ..token import (  # noqa: F401
    Token,
    TokenType,
)


def _is(expected_type, token):
    # type: (TokenType, Optional[Token]) -> bool
    assert token is not None
    return token.token_type == expected_type


class ParserException(BaseException):
    """The exception raised when there is a parsing problem."""

    def __init__(self,
                 msg='',
                 style_error=GenericSyntaxError,
                 line_numbers=None):
        # type: (str, Any, Tuple[int, int]) -> None
        """Create a new ParserException.

        Args:
            msg: The message this error should display.
            style_error: If style errors are supported, then this
                is the type of style error.
            line_numbers: Th eline number where this error occurred.

        """
        super(ParserException, self).__init__(msg)
        self.style_error = style_error
        self.line_numbers = line_numbers


def Assert(expr, msg, style_error=None, token=None):
    # type: (bool, str, Any, Token) -> None
    """Assert that the expression is True.

    Args:
        expr: The expression which we are evaluating.
        msg: The message which will be passed to the error
            if we are wrong.
        style_error: The error to pass into the ParserException
            if `expr` is false.
        token: The last token which was parsed before or at
            this check. (Used for passing along line numbers.)

    Raises:
        ParserException: If `expr` is not true and `style_error`
            is not specified.

    """
    if not expr:
        if token is not None:
            line_numbers = (token.line_number, token.line_number)  # type: Optional[Tuple[int, int]] # noqa
        else:
            line_numbers = None
        if style_error:
            raise ParserException(
                msg,
                style_error=style_error,
                line_numbers=line_numbers,
            )
        else:
            raise ParserException(
                msg,
                line_numbers=line_numbers,
            )


def AssertNotEmpty(peaker, context, style_error=None):
    # type: (Peaker, str, Any) -> None
    """Raise a parser exception if the next item is empty.

    Raises:
        ParserException: If the peaker is, in fact, empty.

    Args:
        peaker: The Peaker which should not be empty.
        context: A verb in the gerund form which describes
            our current actions.
        style_error: The style error passed with the
            ParserException if the Peaker is empty.

    """
    if not peaker.has_next():
        prev_line_numbers = None
        if peaker.prev and peaker.prev.line_number:
            line_no = peaker.prev.line_number
            prev_line_numbers = (line_no, line_no)
        if style_error:
            raise ParserException(
                'Unable to {}: stream was unexpectedly empty.'.format(
                    context
                ),
                style_error=style_error,
                line_numbers=prev_line_numbers,
            )
        else:
            raise ParserException(
                'Unable to {}: stream was unexpectedly empty.'.format(
                    context,
                ),
                line_numbers=prev_line_numbers,
            )


def parse_colon(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse colon')
    Assert(
        _is(TokenType.COLON, peaker.peak()),
        'Unable to parse colon: expected {} but received {}'.format(
            TokenType.COLON, peaker.rpeak().token_type
        ),
        token=peaker.peak()
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.COLON,
        value=token.value,
        token=token,
    )


def parse_word(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse word')
    Assert(
        _is(TokenType.WORD, peaker.peak()),
        'Unable to parse word: expected {} but received {}'.format(
            TokenType.WORD, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.WORD,
        value=token.value,
        token=token,
    )


def parse_hash(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse hash')
    Assert(
        _is(TokenType.HASH, peaker.peak()),
        'Unable to parse hash: expected {} but received {}'.format(
            TokenType.HASH, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.HASH,
        value=token.value,
        token=token,
    )


def parse_lparen(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse left parenthesis')
    Assert(
        _is(TokenType.LPAREN, peaker.peak()),
        'Unable to parse left parenthesis: expected {} '
        'but received {}'.format(
            TokenType.LPAREN, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.LPAREN,
        value=token.value,
        token=token,
    )


def parse_rparen(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse right parenthesis')
    Assert(
        _is(TokenType.RPAREN, peaker.peak()),
        'Unable to parse right parenthesis: expected {} '
        'but received {}'.format(
            TokenType.RPAREN, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.RPAREN,
        value=token.value,
        token=token,
    )


def parse_indent(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse indent')
    Assert(
        _is(TokenType.INDENT, peaker.peak()),
        'Unable to parse indent: expected {} but received {}'.format(
            TokenType.INDENT, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.INDENT,
        value=token.value,
        token=token,
    )


def parse_noqa_head(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_hash(peaker),
    ]
    word = parse_word(peaker)
    Assert(
        word.value == 'noqa',
        'Failed to parse noqa statement. '
        'Expected "# noqa" but received "# {}"'.format(
            word.value,
        ),
        token=peaker.peak(),
    )
    children.append(word)
    return Node(
        node_type=NodeType.NOQA_HEAD,
        children=children,
    )


def parse_noqa_body(peaker):
    # type: (Peaker[Token]) -> Node
    children = [parse_word(peaker)]
    if peaker.has_next() and _is(TokenType.WORD, peaker.peak()):
        children.append(parse_list(peaker))
    return Node(
        node_type=NodeType.NOQA_BODY,
        children=children,
    )


def parse_list(peaker):
    # type: (Peaker[Token]) -> Node
    prev = parse_word(peaker)
    children = [prev]
    while ((prev.value or '').endswith(',')
            and peaker.has_next()
            and _is(TokenType.WORD, peaker.peak())):
        prev = parse_word(peaker)
        children.append(prev)
    return Node(
        node_type=NodeType.LIST,
        children=children
    )


def parse_noqa(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_noqa_head(peaker),
    ]
    if peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        children.extend([
            parse_colon(peaker),
            parse_noqa_body(peaker),
        ])
    return Node(
        node_type=NodeType.NOQA,
        children=children,
    )


def parse_keyword(peaker, keywords=dict()):
    # type: (Peaker[Token], Dict[str, NodeType]) -> Node
    """Parse a keyword.

    Args:
        peaker: A stream of tokens from lexing a docstring.
        keywords: A map of keywords to their associated
            NodeTypes.

    Returns:
        A Node with Keyword NodeType.

    """
    AssertNotEmpty(peaker, 'parse keyword')
    Assert(
        _is(TokenType.WORD, peaker.peak()),
        'Unable to parse keyword: expected {} but received {}'.format(
            TokenType.WORD, peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    Assert(
        peaker.rpeak().value in keywords,
        'Unable to parse keyword: "{}" is not a keyword'.format(
            peaker.rpeak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(keywords[token.value], value=token.value, token=token)



r"""
__EBNF for a google-style docstring__:

  docstring = short-description
              | short-description
                  , newline
                  , newline
                  , {long-description}
                  , [sections];

  short-description = word, {hash | noqa | word | colon | keyword}
  long-description  = head-line, {head-line}
  head-line = indent
                  , [word | colon | indent]
                  , {word | colon | indent | keyword}
                  , [noqa]
                  , newline

  sections = [arguments-section]
                  , [raises-section]
                  , [yields-section | returns-section]
              | [raises-section]
                  , [arguments-section]
                  , [yields-section | returns-section]

  arguments-section = section-compound
  raises-section = section-compound
  yields-section = section-simple
  returns-section = section-simple

  section-simple = section-head, section-simple-body
  section-compound = section-head, section-compound-body
  section-head = keyword, colon, newline
  section-simple-body =
    indent, [type], {word | colon | indent | keyword}, newline
    {indent, indent, line}
  section-compound-body = item, {item}
  item = indent, item-name, colon, item-definition
  item-name = word, [type]
  item-definition = line, {line}
  line = { word
         | hash
         | colon
         | indent
         | keyword
         }, [noqa] newline
  type = lparen, word, {word}, rparen
           | word, colon

  noqa      = noqa-head, [colon, noqa-body]
  noqa-head = hash, word
  noqa-body = word, [list]
  list      = word, {"," word}

  hash    = "#"
  lparen  = "("
  rparen  = ")"
  keyword = "Args"
            | "Arguments"
            | "Returns
            | "Yields"
            | "Raises"
  indent  = " "{4}
  word    = [^\ \n\:\"\t]+
  colon   = ":"
  newline = "\n"

"""
from typing import (  # noqa: F401
    List,
    Optional,
    Set,
)

from ..peaker import Peaker  # noqa: F401
from ..token import Token, TokenType  # noqa: F401
from ..node import Node, NodeType
from ..errors import (
    EmptyDescriptionError,
)
from .common import (
    Assert,
    AssertNotEmpty,
    ParserException,
    parse_colon,
    parse_indent,
    parse_keyword,
    parse_lparen,
    parse_noqa,
    parse_rparen,
    parse_word,
)

KEYWORDS = {
    'Args': NodeType.ARGUMENTS,
    'Arguments': NodeType.ARGUMENTS,
    'Returns': NodeType.RETURNS,
    'Yields': NodeType.YIELDS,
    'Raises': NodeType.RAISES,
}


def _is(expected_type, token):
    # type: (TokenType, Optional[Token]) -> bool
    if token is None:
        return False
    return token.token_type == expected_type


def parse_parenthetical_type(peaker):
    # type: (Peaker[Token]) -> Node
    children = [parse_lparen(peaker)]
    parentheses_count = 1
    encountered_word = False
    while peaker.has_next() and parentheses_count > 0:
        Assert(
            peaker.has_next(),
            'Encountered end of stream while parsing '
            'parenthetical type. Ended with {}'.format(
                [x.value for x in children]
            ),
            token=peaker.peak(),
        )
        if _is(TokenType.LPAREN, peaker.peak()):
            parentheses_count += 1
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, peaker.peak()):
            parentheses_count -= 1
            children.append(parse_rparen(peaker))
        elif _is(TokenType.COLON, peaker.peak()):
            children.append(parse_colon(peaker))
        elif _is(TokenType.INDENT, peaker.peak()):
            parse_indent(peaker)
        else:
            encountered_word = True
            children.append(parse_word(peaker))
    Assert(
        parentheses_count == 0,
        'Mismatched parentheses in parenthetical type.',
        token=peaker.peak(),
    )
    Assert(
        encountered_word,
        'Parenthetical type must contain at least one word.',
        token=peaker.peak(),
    )
    return Node(
        node_type=NodeType.TYPE,
        children=children,
    )


def parse_type(peaker):
    # type: (Peaker[Token]) -> Node
    if _is(TokenType.LPAREN, peaker.peak()):
        return parse_parenthetical_type(peaker)

    AssertNotEmpty(peaker, 'parse type')
    node = parse_word(peaker)
    Assert(
        _is(TokenType.COLON, peaker.peak()),
        'Expected type, "{}", to have "(" and ")" around it or '
        'end in colon.'.format(
            node.value
        ),
        token=peaker.peak(),
    )
    peaker.next()  # Toss the colon
    return Node(
        node_type=NodeType.TYPE,
        children=[node],
    )


def parse_line(peaker, with_type=False):
    # type: (Peaker[Token], bool) -> Node
    AssertNotEmpty(peaker, 'parse line')
    children = list()

    # Get the first node, which may be a type description.
    if with_type and _is(TokenType.WORD, peaker.peak()):
        next_value = peaker.next()
        if next_value.value.startswith('(') and next_value.value.endswith(')'):
            first_node = parse_type(
                Peaker((x for x in [next_value]))
            )
        elif _is(TokenType.COLON, peaker.peak()):
            first_node = parse_type(
                Peaker((x for x in [next_value, peaker.next()]))
            )
        else:
            first_node = parse_word(
                Peaker((x for x in [next_value]))
            )
        children.append(first_node)

    # Get the remaining nodes in the line, up to the newline.
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        next_child = peaker.peak()
        assert next_child is not None
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker, KEYWORDS))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        elif _is(TokenType.HASH, next_child):
            children.append(parse_noqa(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                ),
                line_numbers=(
                    next_child.line_number,
                    next_child.line_number,
                )
            )

    # It is possible that there are no children at this point, in
    # which case there is likely just the newline.  In this case,
    # we try to set the token so that the line can have a line
    # number.
    token = None
    if peaker.has_next():
        if not children:
            token = peaker.next()
        else:
            peaker.next()  # Throw away newline.
    return Node(
        NodeType.LINE,
        children=children,
        token=token,
    )

# NOTE: If Peaker ever allows 2-constant look-ahead, then change
# this to call to the `parse_line(peaker)` function to prevent
# drift between these two functions.


def parse_line_with_type(peaker):
    # type: (Peaker[Token]) -> Node
    """Parse a line which begins with a type description.

    Such lines occur at the start of the yields and returns
    sections.

    Args:
        peaker: A stream of tokens.

    Raises:
        ParserException: If there was any problem with parsing
            the line.

    Returns:
        A line node.

    """
    AssertNotEmpty(peaker, 'parse line')
    children = [
        parse_indent(peaker),
    ]

    # Get the first node, which may be a type description.
    if _is(TokenType.WORD, peaker.peak()):
        next_value = peaker.next()
        if next_value.value.startswith('(') and next_value.value.endswith(')'):
            AssertNotEmpty(peaker, 'parse line')
            first_node = parse_type(
                Peaker((x for x in [next_value]))
            )
        elif _is(TokenType.COLON, peaker.peak()):
            AssertNotEmpty(peaker, 'parse line')
            first_node = parse_type(
                Peaker((x for x in [next_value, peaker.next()]))
            )
        else:
            first_node = parse_word(
                Peaker((x for x in [next_value]))
            )
        children.append(first_node)

    # Get the remaining nodes in the line, up to the newline.
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        next_child = peaker.peak()
        assert next_child is not None, 'Make the type checker happy.'
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker, KEYWORDS))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                )
            )
    return Node(
        NodeType.LINE,
        children=children,
    )


def parse_section_head(peaker, expecting=set()):
    # type: (Peaker[Token], Set[str]) -> Node
    AssertNotEmpty(peaker, 'parse section head')
    children = list()  # type: List[Node]
    # TODO: This error message is too generic; try to make it more specific.
    child = peaker.peak()
    assert child is not None
    Assert(
        child.value in expecting,
        'Expected section head to start with one of {} but was {}'.format(
            expecting,
            repr(child.value),
        ),
        token=peaker.peak(),
    )
    children.append(parse_keyword(peaker, KEYWORDS))
    children.append(parse_colon(peaker))
    # TODO: Allow a noqa here.
    node = peaker.peak()
    Assert(
        _is(TokenType.NEWLINE, node),
        'Failed to parse section head: expected '
        'it to end with {} but encountered {}'.format(
            TokenType.NEWLINE,
            'None' if node is None else node.token_type,
        ),
        token=peaker.peak(),
    )
    peaker.next()
    return Node(
        NodeType.SECTION_HEAD,
        children=children,
    )


def parse_section_simple_body(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse section body')
    children = [
        parse_line_with_type(peaker),
    ]
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        children.append(parse_indent(peaker))
        children.append(parse_line(peaker))
    return Node(
        NodeType.SECTION_SIMPLE_BODY,
        children=children,
    )


def parse_simple_section(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse section')
    children = [
        parse_section_head(peaker, expecting={'Returns', 'Yields'}),
        parse_section_simple_body(peaker),
    ]
    if peaker.has_next():
        Assert(
            _is(TokenType.NEWLINE, peaker.peak()),
            'Expected newline after section.',
            token=peaker.peak(),
        )
        peaker.next()  # Discard newline.
    return Node(
        NodeType.SECTION,
        children=children,
    )


def parse_yields(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_simple_section(peaker)
    node.node_type = NodeType.YIELDS_SECTION
    return node


def parse_returns(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_simple_section(peaker)
    node.node_type = NodeType.RETURNS_SECTION
    return node


def parse_item_name(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse item')
    children = [
        parse_word(peaker),
    ]
    if peaker.has_next() and not _is(TokenType.COLON, peaker.peak()):
        children.append(parse_type(peaker))
    return Node(
        NodeType.ITEM_NAME,
        children=children,
    )


def parse_item_definition(peaker, prev_line_number=None):
    # type: (Peaker[Token], Optional[int]) -> Node

    def _is_indent(i):
        return _is(TokenType.INDENT, peaker.peak(lookahead=i))

    AssertNotEmpty(
        peaker,
        'parse item definition',
        style_error=EmptyDescriptionError
    )
    Assert(
        not _is(TokenType.NEWLINE, peaker.peak()),
        'Unable to parse item definition: stream was unexpectedly empty.',
        token=peaker.peak(),
    )
    children = [
        parse_line(peaker),
    ]
    while _is_indent(1) and _is_indent(2):
        children.append(parse_line(peaker))
    return Node(
        NodeType.ITEM_DEFINITION,
        children=children,
    )


def parse_item(peaker):
    # type: (Peaker[Token]) -> Optional[Node]
    children = [
        parse_indent(peaker),
    ]

    excess_whitespace = False
    while _is(TokenType.INDENT, peaker.peak()):
        excess_whitespace = True
        peaker.next()
    if _is(TokenType.NEWLINE, peaker.peak()):
        return None
    Assert(
        not excess_whitespace,
        'Expected item but encountered extra whitespace.',
    )

    children.append(
        parse_item_name(peaker),
    )

    if _is(TokenType.LPAREN, peaker.peak()):
        children.append(parse_type(peaker))

    children.extend([
        parse_colon(peaker),
        parse_item_definition(peaker),
    ])
    return Node(
        NodeType.ITEM,
        children=children,
    )


def parse_section_compound_body(peaker):
    children = [
        parse_item(peaker),
    ]
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        item = parse_item(peaker)
        whitespace_only_line = item is None
        if whitespace_only_line:
            continue
        children.append(item)

    return Node(
        node_type=NodeType.SECTION_COMPOUND_BODY,
        children=children,
    )


def parse_compound_section(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_section_head(peaker, expecting={'Args', 'Arguments', 'Raises'}),
        parse_section_compound_body(peaker),
    ]
    if peaker.has_next():
        node = peaker.peak()
        assert node is not None
        Assert(
            _is(TokenType.NEWLINE, node),
            'Expected {} after compound section but received {}'.format(
                TokenType.NEWLINE,
                node.token_type
            ),
            token=peaker.peak(),
        )
        peaker.next()  # discard newline.

    return Node(
        node_type=NodeType.SECTION,
        children=children,
    )


def parse_args(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_compound_section(peaker)
    node.node_type = NodeType.ARGS_SECTION
    return node


def parse_raises(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_compound_section(peaker)
    node.node_type = NodeType.RAISES_SECTION
    return node


def parse_short_description(peaker):
    # type: (Peaker[Token]) -> Node
    children = list()
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        next_child = peaker.peak()
        assert next_child is not None
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker, KEYWORDS))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        elif _is(TokenType.HASH, next_child):
            children.append(parse_noqa(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                )
            )
    if peaker.has_next() and _is(TokenType.NEWLINE, peaker.peak()):
        peaker.next()  # Eat the newline.
    return Node(
        node_type=NodeType.SHORT_DESCRIPTION,
        children=children,
    )


def parse_long_description(peaker):
    # type: (Peaker[Token]) -> Node
    node = peaker.peak()
    Assert(
        node is not None and node.value not in KEYWORDS,
        'Expected long description to start with non-keyword but {}.'.format(
            'was empty.' if not peaker.has_next() else 'was {}'.format(
                'None' if node is None else node.value
            )
        ),
        token=peaker.peak(),
    )
    children = [
        parse_line(peaker),
    ]
    node = peaker.peak()
    while node is not None and node.value not in KEYWORDS:
        children.append(parse_line(peaker))
        node = peaker.peak()

    return Node(
        node_type=NodeType.LONG_DESCRIPTION,
        children=children,
    )


def parse_description(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_short_description(peaker),
    ]
    if peaker.has_next():
        node = peaker.peak()
        assert node is not None
        Assert(
            _is(TokenType.NEWLINE, node),
            'Expected blank line after short description, but '
            'found {}: {}.'.format(
                node.token_type,
                repr(node.value)
            ),
            token=peaker.peak(),
        )
        peaker.next()  # consume blank line.
    node = peaker.peak()
    if node is not None and node.value not in KEYWORDS:
        children.append(parse_long_description(peaker))
    return Node(
        node_type=NodeType.DESCRIPTION,
        children=children,
    )


def parse(peaker):
    # type: (Peaker[Token]) -> Node
    """Parse the docstring.

    Args:
        peaker: A Peaker filled with the lexed tokens of the
            docstring.

    Raises:
        ParserException: If there is anything malformed with
            the docstring, or if anything goes wrong with parsing. # noqa

    Returns:
        The parsed docstring as an AST.

    """
    keyword_parse_lookup = {
        'Args': parse_args,
        'Arguments': parse_args,
        'Returns': parse_returns,
        'Yields': parse_yields,
        'Raises': parse_raises,
    }
    children = [
        parse_description(peaker)
    ]
    while peaker.has_next():
        node = peaker.peak()
        assert node is not None
        next_value = node.value
        if next_value in keyword_parse_lookup:
            children.append(
                keyword_parse_lookup[next_value](peaker)
            )
        else:
            children.append(
                parse_long_description(peaker)
            )
    return Node(
        node_type=NodeType.DOCSTRING,
        children=children,
    )



r"""
__EBNF for a google-style docstring__:

  <docstring> ::= <short-description>
                | <short-description><newline><newline>
                    <long-description>*
                    <sections>*

  <short-description> ::= <word>[<hash><noqa><word><colon><keyword>]*
  <long-description>  ::= <head-line>+
  <head-line> ::= <indent>
                    [<word><colon><indent>]
                    [<word><colon><indent><keyword>]*
                    <noqa>?<newline>

  <sections> ::= <arguments-section>?
                   <raises-section>?
                   (<yields-section>|<returns-section>)?
               | <raises-section>?
                   <arguments-section>?
                   (yields-section>|<returns-section>)?

  <arguments-section> ::= <section-compound>
  <raises-section> ::= <section-compound>
  <yields-section> ::= <section-simple>
  <returns-section> ::= <section-simple>

  <section-simple> ::= <section-head><section-simple-body>
  <section-compound> ::= <section-head><section-compound-body>
  <section-head> ::= <keyword><colon><newline>
  <section-simple-body> ::=
    <indent><type>?[<word><colon><indent><keyword>]*<newline>
    (<indent><indent><line>)*
  <section-compound-body> ::= <item>+
  <item> ::= <indent><item-name><colon><item-definition>
  <item-name> ::= <word><type>?
  <item-definition> ::= <line>+
  <line> ::= [<word><hash><colon><indent><keyword>]*<noqa>?<newline>
  <type> ::= <lparen><word>+<rparen>
           | <word><colon>

  <noqa> ::= <noqa-head>(<colon><noqa-body>)?
  <noqa-head> ::= <hash><word>
  <noqa-body> ::= <word><list>?
  <list> ::= <word>["," <word>]*

  <hash> ::= "#"
  <lparen> ::= "("
  <rparen> ::= ")"
  <keyword> ::= "Args"
            | "Arguments"
            | "Returns
            | "Yields"
            | "Raises"
  <indent>  ::= " "{4}
  <word>    ::= [^\ \n\:\"\t]+
  <colon>   ::= ":"
  <newline> ::= "\n"

"""
from typing import (
    Any,
    List,
    Optional,
    Set,
    Tuple,
)  # noqa

from .peaker import Peaker  # noqa
from .token import Token, TokenType  # noqa
from .node import Node, NodeType
from .errors import (
    GenericSyntaxError,
    EmptyDescriptionError,
)

KEYWORDS = {
    'Args': NodeType.ARGUMENTS,
    'Arguments': NodeType.ARGUMENTS,
    'Returns': NodeType.RETURNS,
    'Yields': NodeType.YIELDS,
    'Raises': NodeType.RAISES,
}


class ParserException(BaseException):
    """The exception raised when there is a parsing problem."""

    def __init__(self,
                 msg='',
                 style_error=GenericSyntaxError,
                 line_numbers=None):
        # type: (str, Any, Tuple[int, int]) -> None
        """Create a new ParserException.

        Args:
            msg: The message this error should display.
            style_error: If style errors are supported, then this
                is the type of style error.
            line_numbers: Th eline number where this error occurred.

        """
        super(ParserException, self).__init__(msg)
        self.style_error = style_error
        self.line_numbers = line_numbers


def Assert(expr, msg, style_error=None, token=None):
    # type: (bool, str, Any, Token) -> None
    """Assert that the expression is True.

    Args:
        expr: The expression which we are evaluating.
        msg: The message which will be passed to the error
            if we are wrong.
        style_error: The error to pass into the ParserException
            if `expr` is false.
        token: The last token which was parsed before or at
            this check. (Used for passing along line numbers.)

    Raises:
        ParserException: If `expr` is not true and `style_error`
            is not specified.

    """
    if not expr:
        if token is not None:
            line_numbers = (token.line_number, token.line_number)  # type: Optional[Tuple[int, int]] # noqa
        else:
            line_numbers = None
        if style_error:
            raise ParserException(
                msg,
                style_error=style_error,
                line_numbers=line_numbers,
            )
        else:
            raise ParserException(
                msg,
                line_numbers=line_numbers,
            )


def AssertNotEmpty(peaker, context, style_error=None):
    # type: (Peaker, str, Any) -> None
    """Raise a parser exception if the next item is empty.

    Raises:
        ParserException: If the peaker is, in fact, empty.

    Args:
        peaker: The Peaker which should not be empty.
        context: A verb in the gerund form which describes
            our current actions.
        style_error: The style error passed with the
            ParserException if the Peaker is empty.

    """
    if not peaker.has_next():
        prev_line_numbers = None
        if peaker.prev and peaker.prev.line_number:
            line_no = peaker.prev.line_number
            prev_line_numbers = (line_no, line_no)
        if style_error:
            raise ParserException(
                'Unable to {}: stream was unexpectedly empty.'.format(
                    context
                ),
                style_error=style_error,
                line_numbers=prev_line_numbers,
            )
        else:
            raise ParserException(
                'Unable to {}: stream was unexpectedly empty.'.format(
                    context,
                ),
                line_numbers=prev_line_numbers,
            )


def _is(expected_type, token):
    # type: (TokenType, Token) -> bool
    return token.token_type == expected_type


def parse_keyword(peaker):
    # type: (Peaker[Token]) -> Node
    """Parse a keyword.

    Args:
        peaker: A stream of tokens from lexing a docstring.

    Returns:
        A Node with Keyword NodeType.

    """
    AssertNotEmpty(peaker, 'parse keyword')
    Assert(
        _is(TokenType.WORD, peaker.peak()),
        'Unable to parse keyword: expected {} but received {}'.format(
            TokenType.WORD, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    Assert(
        peaker.peak().value in KEYWORDS,
        'Unable to parse keyword: "{}" is not a keyword'.format(
            peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(KEYWORDS[token.value], value=token.value, token=token)


def parse_colon(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse colon')
    Assert(
        _is(TokenType.COLON, peaker.peak()),
        'Unable to parse colon: expected {} but received {}'.format(
            TokenType.COLON, peaker.peak().token_type
        ),
        token=peaker.peak()
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.COLON,
        value=token.value,
        token=token,
    )


def parse_word(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse word')
    Assert(
        _is(TokenType.WORD, peaker.peak()),
        'Unable to parse word: expected {} but received {}'.format(
            TokenType.WORD, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.WORD,
        value=token.value,
        token=token,
    )


def parse_hash(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse hash')
    Assert(
        _is(TokenType.HASH, peaker.peak()),
        'Unable to parse hash: expected {} but received {}'.format(
            TokenType.HASH, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.HASH,
        value=token.value,
        token=token,
    )


def parse_lparen(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse left parenthesis')
    Assert(
        _is(TokenType.LPAREN, peaker.peak()),
        'Unable to parse left parenthesis: expected {} '
        'but received {}'.format(
            TokenType.LPAREN, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.LPAREN,
        value=token.value,
        token=token,
    )


def parse_rparen(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse right parenthesis')
    Assert(
        _is(TokenType.RPAREN, peaker.peak()),
        'Unable to parse right parenthesis: expected {} '
        'but received {}'.format(
            TokenType.RPAREN, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.RPAREN,
        value=token.value,
        token=token,
    )


def parse_parenthetical_type(peaker):
    # type: (Peaker[Token]) -> Node
    children = [parse_lparen(peaker)]
    parentheses_count = 1
    encountered_word = False
    while peaker.has_next() and parentheses_count > 0:
        Assert(
            peaker.has_next(),
            'Encountered end of stream while parsing '
            'parenthetical type. Ended with {}'.format(
                [x.value for x in children]
            ),
            token=peaker.peak(),
        )
        if _is(TokenType.LPAREN, peaker.peak()):
            parentheses_count += 1
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, peaker.peak()):
            parentheses_count -= 1
            children.append(parse_rparen(peaker))
        elif _is(TokenType.COLON, peaker.peak()):
            children.append(parse_colon(peaker))
        else:
            encountered_word = True
            children.append(parse_word(peaker))
    Assert(
        parentheses_count == 0,
        'Mismatched parentheses in parenthetical type.',
        token=peaker.peak(),
    )
    Assert(
        encountered_word,
        'Parenthetical type must contain at least one word.',
        token=peaker.peak(),
    )
    return Node(
        node_type=NodeType.TYPE,
        children=children,
    )


def parse_type(peaker):
    # type: (Peaker[Token]) -> Node
    if _is(TokenType.LPAREN, peaker.peak()):
        return parse_parenthetical_type(peaker)

    AssertNotEmpty(peaker, 'parse type')
    node = parse_word(peaker)
    Assert(
        _is(TokenType.COLON, peaker.peak()),
        'Expected type, "{}", to have "(" and ")" around it or '
        'end in colon.'.format(
            node.value
        ),
        token=peaker.peak(),
    )
    peaker.next()  # Toss the colon
    return Node(
        node_type=NodeType.TYPE,
        children=[node],
    )


def parse_indent(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse indent')
    Assert(
        _is(TokenType.INDENT, peaker.peak()),
        'Unable to parse indent: expected {} but received {}'.format(
            TokenType.INDENT, peaker.peak().token_type
        ),
        token=peaker.peak(),
    )
    token = peaker.next()
    return Node(
        node_type=NodeType.INDENT,
        value=token.value,
        token=token,
    )


def parse_line(peaker, with_type=False):
    # type: (Peaker[Token], bool) -> Node
    AssertNotEmpty(peaker, 'parse line')
    children = list()

    # Get the first node, which may be a type description.
    if with_type and _is(TokenType.WORD, peaker.peak()):
        next_value = peaker.next()
        if next_value.value.startswith('(') and next_value.value.endswith(')'):
            first_node = parse_type(
                Peaker((x for x in [next_value]))
            )
        elif _is(TokenType.COLON, peaker.peak()):
            first_node = parse_type(
                Peaker((x for x in [next_value, peaker.next()]))
            )
        else:
            first_node = parse_word(
                Peaker((x for x in [next_value]))
            )
        children.append(first_node)

    # Get the remaining nodes in the line, up to the newline.
    while peaker.has_next() and not peaker.peak().token_type == TokenType.NEWLINE:
        next_child = peaker.peak()
        assert next_child is not None
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        elif _is(TokenType.HASH, next_child):
            children.append(parse_noqa(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                ),
                line_numbers=(
                    next_child.line_number,
                    next_child.line_number,
                )
            )

    # It is possible that there are no children at this point, in
    # which case there is likely just the newline.  In this case,
    # we try to set the token so that the line can have a line
    # number.
    token = None
    if peaker.has_next():
        if not children:
            token = peaker.next()
        else:
            peaker.next()  # Throw away newline.
    return Node(
        NodeType.LINE,
        children=children,
        token=token,
    )

# NOTE: If Peaker ever allows 2-constant look-ahead, then change
# this to call to the `parse_line(peaker)` function to prevent
# drift between these two functions.


def parse_line_with_type(peaker):
    # type: (Peaker[Token]) -> Node
    """Parse a line which begins with a type description.

    Such lines occur at the start of the yields and returns
    sections.

    Args:
        peaker: A stream of tokens.

    Raises:
        ParserException: If there was any problem with parsing
            the line.

    Returns:
        A line node.

    """
    AssertNotEmpty(peaker, 'parse line')
    children = [
        parse_indent(peaker),
    ]

    # Get the first node, which may be a type description.
    if _is(TokenType.WORD, peaker.peak()):
        next_value = peaker.next()
        AssertNotEmpty(peaker, 'parse line')
        if next_value.value.startswith('(') and next_value.value.endswith(')'):
            first_node = parse_type(
                Peaker((x for x in [next_value]))
            )
        elif _is(TokenType.COLON, peaker.peak()):
            first_node = parse_type(
                Peaker((x for x in [next_value, peaker.next()]))
            )
        else:
            first_node = parse_word(
                Peaker((x for x in [next_value]))
            )
        children.append(first_node)

    # Get the remaining nodes in the line, up to the newline.
    while (peaker.has_next()
            and not peaker.peak().token_type == TokenType.NEWLINE):
        next_child = peaker.peak()
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                )
            )
    return Node(
        NodeType.LINE,
        children=children,
    )


def parse_section_head(peaker, expecting=set()):
    # type: (Peaker[Token], Set[str]) -> Node
    AssertNotEmpty(peaker, 'parse section head')
    children = list()  # type: List[Node]
    # TODO: This error message is too generic; try to make it more specific.
    Assert(
        peaker.peak().value in expecting,
        'Expected section head to start with one of {} but was {}'.format(
            expecting,
            repr(peaker.peak().value),
        ),
        token=peaker.peak(),
    )
    children.append(parse_keyword(peaker))
    children.append(parse_colon(peaker))
    # TODO: Allow a noqa here.
    Assert(
        _is(TokenType.NEWLINE, peaker.peak()),
        'Failed to parse section head: expected '
        'it to end with {} but encountered {}'.format(
            TokenType.NEWLINE,
            peaker.peak().token_type,
        ),
        token=peaker.peak(),
    )
    peaker.next()
    return Node(
        NodeType.SECTION_HEAD,
        children=children,
    )


def parse_section_simple_body(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse section body')
    children = [
        parse_line_with_type(peaker),
    ]
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        children.append(parse_indent(peaker))
        children.append(parse_line(peaker))
    return Node(
        NodeType.SECTION_SIMPLE_BODY,
        children=children,
    )


def parse_simple_section(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse section')
    children = [
        parse_section_head(peaker, expecting={'Returns', 'Yields'}),
        parse_section_simple_body(peaker),
    ]
    if peaker.has_next():
        Assert(
            _is(TokenType.NEWLINE, peaker.peak()),
            'Expected newline after section.',
            token=peaker.peak(),
        )
        peaker.next()  # Discard newline.
    return Node(
        NodeType.SECTION,
        children=children,
    )


def parse_yields(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_simple_section(peaker)
    node.node_type = NodeType.YIELDS_SECTION
    return node


def parse_returns(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_simple_section(peaker)
    node.node_type = NodeType.RETURNS_SECTION
    return node


def parse_item_name(peaker):
    # type: (Peaker[Token]) -> Node
    AssertNotEmpty(peaker, 'parse item')
    children = [
        parse_word(peaker),
    ]
    if peaker.has_next() and not _is(TokenType.COLON, peaker.peak()):
        children.append(parse_type(peaker))
    return Node(
        NodeType.ITEM_NAME,
        children=children,
    )


def parse_item_definition(peaker, prev_line_number=None):
    # type: (Peaker[Token], Optional[int]) -> Node

    def _is_indent(i):
        token = peaker.peak(lookahead=i)
        return token is not None and _is(TokenType.INDENT, token)

    AssertNotEmpty(
        peaker,
        'parse item definition',
        style_error=EmptyDescriptionError
    )
    Assert(
        not _is(TokenType.NEWLINE, peaker.peak()),
        'Unable to parse item definition: stream was unexpectedly empty.',
        token=peaker.peak(),
    )
    children = [
        parse_line(peaker),
    ]
    while _is_indent(1) and _is_indent(2):
        children.append(parse_line(peaker))
    return Node(
        NodeType.ITEM_DEFINITION,
        children=children,
    )


def parse_item(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_indent(peaker),
        parse_item_name(peaker),
    ]

    if _is(TokenType.LPAREN, peaker.peak()):
        children.append(parse_type(peaker))

    children.extend([
        parse_colon(peaker),
        parse_item_definition(peaker),
    ])
    return Node(
        NodeType.ITEM,
        children=children,
    )


def parse_section_compound_body(peaker):
    children = [
        parse_item(peaker),
    ]
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        children.append(parse_item(peaker))

    return Node(
        node_type=NodeType.SECTION_COMPOUND_BODY,
        children=children,
    )


def parse_compound_section(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_section_head(peaker, expecting={'Args', 'Arguments', 'Raises'}),
        parse_section_compound_body(peaker),
    ]
    if peaker.has_next():
        Assert(
            _is(TokenType.NEWLINE, peaker.peak()),
            'Expected {} after compound section but received {}'.format(
                TokenType.NEWLINE,
                peaker.peak().token_type
            ),
            token=peaker.peak(),
        )
        peaker.next()  # discard newline.

    return Node(
        node_type=NodeType.SECTION,
        children=children,
    )


def parse_args(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_compound_section(peaker)
    node.node_type = NodeType.ARGS_SECTION
    return node


def parse_raises(peaker):
    # type: (Peaker[Token]) -> Node
    node = parse_compound_section(peaker)
    node.node_type = NodeType.RAISES_SECTION
    return node


def parse_short_description(peaker):
    # type: (Peaker[Token]) -> Node
    children = list()
    while peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        next_child = peaker.peak()
        if _is(TokenType.WORD, next_child) and next_child.value in KEYWORDS:
            children.append(parse_keyword(peaker))
        elif _is(TokenType.WORD, next_child):
            children.append(parse_word(peaker))
        elif _is(TokenType.INDENT, next_child):
            children.append(parse_indent(peaker))
        elif _is(TokenType.COLON, next_child):
            children.append(parse_colon(peaker))
        elif _is(TokenType.LPAREN, next_child):
            children.append(parse_lparen(peaker))
        elif _is(TokenType.RPAREN, next_child):
            children.append(parse_rparen(peaker))
        elif _is(TokenType.HASH, next_child):
            children.append(parse_noqa(peaker))
        else:
            raise ParserException(
                'Failed to parse line: invalid token type {}'.format(
                    next_child.token_type
                )
            )
    if peaker.has_next() and _is(TokenType.NEWLINE, peaker.peak()):
        peaker.next()  # Eat the newline.
    return Node(
        node_type=NodeType.SHORT_DESCRIPTION,
        children=children,
    )


def parse_long_description(peaker):
    # type: (Peaker[Token]) -> Node
    Assert(
        peaker.has_next() and peaker.peak().value not in KEYWORDS,
        'Expected long description to start with non-keyword but {}.'.format(
            'was empty.' if not peaker.has_next() else 'was {}'.format(
                peaker.peak().value
            )
        ),
        token=peaker.peak(),
    )
    children = [
        parse_line(peaker),
    ]
    while peaker.has_next() and peaker.peak().value not in KEYWORDS:
        children.append(parse_line(peaker))

    return Node(
        node_type=NodeType.LONG_DESCRIPTION,
        children=children,
    )


def parse_description(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_short_description(peaker),
    ]
    if peaker.has_next():
        Assert(
            _is(TokenType.NEWLINE, peaker.peak()),
            'Expected blank line after short description, but '
            'found {}: {}.'.format(
                peaker.peak().token_type,
                repr(peaker.peak().value)
            ),
            token=peaker.peak(),
        )
        peaker.next()  # consume blank line.
    if peaker.has_next() and peaker.peak().value not in KEYWORDS:
        children.append(parse_long_description(peaker))
    return Node(
        node_type=NodeType.DESCRIPTION,
        children=children,
    )


def parse_noqa_head(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_hash(peaker),
    ]
    word = parse_word(peaker)
    Assert(
        word.value == 'noqa',
        'Failed to parse noqa statement. '
        'Expected "# noqa" but received "# {}"'.format(
            word.value,
        ),
        token=peaker.peak(),
    )
    children.append(word)
    return Node(
        node_type=NodeType.NOQA_HEAD,
        children=children,
    )


def parse_list(peaker):
    # type: (Peaker[Token]) -> Node
    prev = parse_word(peaker)
    children = [prev]
    while (prev.value.endswith(',')
            and peaker.has_next()
            and _is(TokenType.WORD, peaker.peak())):
        prev = parse_word(peaker)
        children.append(prev)
    return Node(
        node_type=NodeType.LIST,
        children=children
    )


def parse_noqa_body(peaker):
    # type: (Peaker[Token]) -> Node
    children = [parse_word(peaker)]
    if peaker.has_next() and _is(TokenType.WORD, peaker.peak()):
        children.append(parse_list(peaker))
    return Node(
        node_type=NodeType.NOQA_BODY,
        children=children,
    )


def parse_noqa(peaker):
    # type: (Peaker[Token]) -> Node
    children = [
        parse_noqa_head(peaker),
    ]
    if peaker.has_next() and not _is(TokenType.NEWLINE, peaker.peak()):
        children.extend([
            parse_colon(peaker),
            parse_noqa_body(peaker),
        ])
    return Node(
        node_type=NodeType.NOQA,
        children=children,
    )


def parse(peaker):
    # type: (Peaker[Token]) -> Node
    """Parse the docstring.

    Args:
        peaker: A Peaker filled with the lexed tokens of the
            docstring.

    Raises:
        ParserException: If there is anything malformed with
            the docstring, or if anything goes wrong with parsing. # noqa

    Returns:
        The parsed docstring as an AST.

    """
    keyword_parse_lookup = {
        'Args': parse_args,
        'Arguments': parse_args,
        'Returns': parse_returns,
        'Yields': parse_yields,
        'Raises': parse_raises,
    }
    children = [
        parse_description(peaker)
    ]
    while peaker.has_next():
        next_value = peaker.peak().value
        if next_value in keyword_parse_lookup:
            children.append(
                keyword_parse_lookup[next_value](peaker)
            )
        else:
            children.append(
                parse_long_description(peaker)
            )
    return Node(
        node_type=NodeType.DOCSTRING,
        children=children,
    )



"""Utilities for debugging Darglint."""

from .node import Node  # noqa: F401


def generate_dot(node):
    # type: (Node) -> str
    """Generate a dot file for the given AST.

    Once generated, the returned tree can be rendered using
    the graphviz library.  For example, to render as an SVG:

        dot -Tsvg rendered_tree.dot -o rendered_tree.svg

    Args:
        node: The root of the tree we are generating.

    Returns:
        The rendered tree, as a dot file.

    """

    def _node_type_str(n):
        return str(n.node_type).replace('NodeType.', '')

    def _get_node_name(n):
        # type: (Node) -> str
        return '{}_{}'.format(
            _node_type_str(n), hash(n)
        ).replace('-', '')

    def _render_node_definition(n):
        # type: (Node) -> str
        name = _get_node_name(n)
        node_is_parent = n.value is None
        if node_is_parent:
            return '{} [label="{}"];'.format(
                name,
                _node_type_str(n),
            )
        else:
            assert n.value is not None
            value = '\\n\\"' + n.value + '\\"'
            return '{} [label="{} {}", shape="rectangle"];'.format(
                name,
                _node_type_str(n),
                value
            )

    def _render_relations(n):
        # type: (Node) -> str
        if not n.children:
            return ''
        ret = '{} -> \t{};'.format(
            _get_node_name(n),
            '\n\t, '.join([_get_node_name(child) for child in n.children])
        )
        for child in [x for x in n.children if x.children]:
            ret += '\n' + _render_relations(child)
        return ret

    definitions = [
        _render_node_definition(n) for n in node.walk()
    ]
    relations = _render_relations(node)

    return '\n\n'.join([
        'digraph G {',
        '\n'.join(definitions),
        relations,
        '}',
    ])
